@@ -14,9 +14,9 @@ gk_dialog (GtkMessageType type, gchar *format, ...)
   msg = g_strdup_vprintf(format, ap);
   va_end(ap);
 
-  diag_win = gtk_message_dialog_new (NULL, GTK_DIALOG_MODAL,
-				     type, GTK_BUTTONS_CLOSE,
-				     msg);
+  diag_win = gtk_message_dialog_new_with_markup (NULL, GTK_DIALOG_MODAL,
+						 type, GTK_BUTTONS_CLOSE,
+						 msg);
 			     
   gtk_signal_connect_object (GTK_OBJECT(diag_win), "delete_event",
 			     GTK_SIGNAL_FUNC(gtk_main_quit), 
